"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from typing import (
    NoReturn,
)

from ...database import (
    PostgreSqlMinosDatabase,
)


class PostgreSqlSnapshotSetup(PostgreSqlMinosDatabase):
    """Minos Snapshot Setup Class"""

    async def _setup(self) -> NoReturn:
        await self.submit_query(_CREATE_TABLE_QUERY, lock=hash("snapshot"))
        await self.submit_query(_CREATE_OFFSET_TABLE_QUERY, lock=hash("snapshot"))


_CREATE_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS snapshot (
    aggregate_uuid UUID NOT NULL,
    aggregate_name TEXT NOT NULL,
    version INT NOT NULL,
    data BYTEA,
    created_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
    updated_at TIMESTAMPTZ NOT NULL DEFAULT NOW(),
    PRIMARY KEY (aggregate_uuid, aggregate_name)
);
""".strip()

_CREATE_OFFSET_TABLE_QUERY = """
CREATE TABLE IF NOT EXISTS snapshot_aux_offset (
    id bool PRIMARY KEY DEFAULT TRUE,
    value BIGINT NOT NULL,
    CONSTRAINT id_uni CHECK (id)
);
""".strip()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .builders import (
    TypeHintBuilder,
    build_union,
)
from .comparators import (
    TypeHintComparator,
    is_aggregate_type,
    is_model_subclass,
    is_type_subclass,
)
from .data_types import (
    Decimal,
    Enum,
    Fixed,
    MissingSentinel,
    NoneType,
)
from .model_refs import (
    ModelRef,
    ModelRefExtractor,
    ModelRefInjector,
)
from .model_types import (
    ModelType,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from asyncio import (
    AbstractEventLoop,
)
from typing import (
    NoReturn,
    Type,
    Union,
)

from aiomisc import (
    Service,
    receiver,
)
from aiomisc.entrypoint import (
    Entrypoint,
)
from aiomisc.log import (
    LogFormat,
    basic_config,
)
from aiomisc.log.enum import (
    DateFormat,
)
from aiomisc.utils import (
    create_default_event_loop,
)
from cached_property import (
    cached_property,
)

from .configuration import (
    MinosConfig,
)
from .importlib import (
    import_module,
)
from .injectors import (
    DependencyInjector,
)
from .setup import (
    MinosSetup,
)

logger = logging.getLogger(__name__)


def _create_entrypoint(*args, **kwargs) -> Entrypoint:  # pragma: no cover
    return Entrypoint(*args, **kwargs)


def _create_loop() -> AbstractEventLoop:  # pragma: no cover
    return create_default_event_loop()[0]


class EntrypointLauncher(MinosSetup):
    """EntryPoint Launcher class."""

    def __init__(
        self,
        config: MinosConfig,
        injections: dict[str, Union[MinosSetup, Type[MinosSetup], str]],
        services: list[Union[Service, Type[Service], str]],
        interval: float = 0.1,
        log_level: Union[int, str] = logging.INFO,
        log_format: Union[str, LogFormat] = "color",
        log_date_format: Union[str, DateFormat] = DateFormat["color"],
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.config = config
        self.interval = interval

        self._log_level = log_level
        self._log_format = log_format
        self._log_date_format = log_date_format

        self._raw_injections = injections
        self._raw_services = services

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> EntrypointLauncher:
        if "injections" not in kwargs:
            kwargs["injections"] = config.service.injections
        if "services" not in kwargs:
            kwargs["services"] = config.service.services
        return cls(config, *args, **kwargs)

    def launch(self) -> NoReturn:
        """Launch a new execution and keeps running forever..

        :return: This method does not return anything.
        """

        basic_config(
            level=self._log_level, log_format=self._log_format, buffered=False, date_format=self._log_date_format,
        )

        logger.info("Starting microservice...")

        try:
            with self.entrypoint:
                logger.info("Microservice is up and running!")
                self.loop.run_forever()
        except KeyboardInterrupt:  # pragma: no cover
            logger.info("Stopping microservice...")
        finally:
            self.graceful_shutdown()

    def graceful_shutdown(self, err: Exception = None) -> NoReturn:
        """Shutdown the services execution gracefully.

        :return: This method does not return anything.
        """
        self.loop.run_until_complete(self.entrypoint.graceful_shutdown(err))

    @cached_property
    def entrypoint(self) -> Entrypoint:
        """Entrypoint instance.

        :return: An ``Entrypoint`` instance.
        """

        # noinspection PyUnusedLocal
        @receiver(Entrypoint.PRE_START)
        async def _start(*args, **kwargs):
            await self.setup()

        # noinspection PyUnusedLocal
        @receiver(Entrypoint.POST_STOP)
        async def _stop(*args, **kwargs):
            await self.destroy()

        return _create_entrypoint(*self.services, loop=self.loop, log_config=False)

    @cached_property
    def loop(self) -> AbstractEventLoop:
        """Create the loop.

        :return: An ``AbstractEventLoop`` instance.
        """
        return _create_loop()

    @cached_property
    def services(self) -> list[Service]:
        """List of services to be launched.

        :return: A list of ``Service`` instances.
        """
        kwargs = {"config": self.config, "interval": self.interval}

        def _fn(raw: Union[Service, Type[Service], str]) -> Service:
            if isinstance(raw, str):
                raw = import_module(raw)
            if isinstance(raw, type):
                return raw(**kwargs)
            return raw

        return [_fn(raw) for raw in self._raw_services]

    async def _setup(self) -> NoReturn:
        """Wire the dependencies and setup it.

        :return: This method does not return anything.
        """
        await self.injector.wire(modules=self._internal_modules)

    @property
    def _internal_modules(self):
        from minos import (
            common,
        )

        modules = [common]
        try:
            # noinspection PyUnresolvedReferences
            from minos import (
                networks,
            )

            modules += [networks]  # pragma: no cover
        except ImportError:
            pass

        try:
            # noinspection PyUnresolvedReferences
            from minos import (
                saga,
            )

            modules += [saga]  # pragma: no cover
        except ImportError:
            pass

        try:
            # noinspection PyUnresolvedReferences
            from minos import (
                cqrs,
            )

            modules += [cqrs]  # pragma: no cover
        except ImportError:
            pass

        return modules

    async def _destroy(self) -> NoReturn:
        """Unwire the injected dependencies and destroys it.

        :return: This method does not return anything.
        """
        await self.injector.unwire()

    @cached_property
    def injector(self) -> DependencyInjector:
        """Dependency injector instance.

        :return: A ``DependencyInjector`` instance.
        """
        return DependencyInjector(config=self.config, **self._raw_injections)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from collections import (
    defaultdict,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    Iterable,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    get_args,
    get_origin,
)
from uuid import (
    UUID,
)

if TYPE_CHECKING:
    from ..abc import (
        Model,
    )

T = TypeVar("T")


class ModelRef(Generic[T]):
    """Model Reference."""


class ModelRefExtractor:
    """Model Reference Extractor class."""

    def __init__(self, value: Any, kind: Optional[Type] = None):
        if kind is None:
            from .builders import (
                TypeHintBuilder,
            )

            kind = TypeHintBuilder(value).build()
        self.value = value
        self.kind = kind

    def build(self) -> dict[str, set[UUID]]:
        """Run the model reference extractor.

        :return: A dictionary in which the keys are the class names and the values are the identifiers.
        """
        ans = defaultdict(set)
        self._build(self.value, self.kind, ans)
        return ans

    def _build(self, value: Any, kind: Type, ans: dict[str, set[UUID]]) -> NoReturn:
        from ..abc import (
            Model,
        )

        if isinstance(value, (tuple, list, set)):
            self._build_iterable(value, get_args(kind)[0], ans)

        elif isinstance(value, dict):
            self._build_iterable(value.keys(), get_args(kind)[0], ans)
            self._build_iterable(value.values(), get_args(kind)[1], ans)

        elif isinstance(value, Model):
            for field in value.fields.values():
                self._build(field.value, field.type, ans)

        elif get_origin(kind) is ModelRef and isinstance(value, UUID):
            cls = get_args(kind)[0]
            name = cls.__name__
            ans[name].add(value)

    def _build_iterable(self, value: Iterable, kind: Type, ans: dict[str, set[UUID]]) -> NoReturn:
        for sub_value in value:
            self._build(sub_value, kind, ans)


class ModelRefInjector:
    """Model Reference Injector class."""

    def __init__(self, value: Any, mapper: dict[UUID, Model]):
        self.value = value
        self.mapper = mapper

    def build(self) -> Any:
        """Inject the model instances referenced by identifiers.

        :return: A model in which the model references have been replaced by the values.
        """
        return self._build(self.value)

    def _build(self, value: Any) -> NoReturn:
        from ..abc import (
            Model,
        )

        if isinstance(value, (tuple, list, set)):
            return type(value)(self._build(v) for v in value)

        if isinstance(value, dict):
            return type(value)((self._build(k), self._build(v)) for k, v in value.items())

        if isinstance(value, Model):
            for field in value.fields.values():
                field.value = self._build(field.value)
            return value

        if isinstance(value, UUID) and value in self.mapper:
            return self.mapper[value]

        return value



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from itertools import (
    zip_longest,
)
from typing import (
    Any,
    Generic,
    NoReturn,
    TypeVar,
    get_type_hints,
)

from ...meta import (
    self_or_classmethod,
)
from ..abc import (
    Model,
)
from ..fields import (
    Field,
)
from ..types import (
    MissingSentinel,
    ModelType,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


class DeclarativeModel(Model, Generic[T]):
    """Base class for ``minos`` declarative model entities."""

    def __init__(self, *args, **kwargs):
        """Class constructor.

        :param kwargs: Named arguments to be set as model attributes.
        """
        super().__init__()
        self._list_fields(*args, **kwargs)

    # noinspection PyUnusedLocal
    @classmethod
    def from_model_type(cls, model_type: ModelType, data: dict[str, Any]):
        """Build a ``DeclarativeModel`` from a ``ModelType`` and ``data``.

        :param model_type: ``ModelType`` object containing the DTO's structure
        :param data: A dictionary containing the values to be stored on the DTO.
        :return: A new ``DeclarativeModel`` instance.
        """
        return cls(**data)

    def _list_fields(self, *args, **kwargs) -> NoReturn:
        for (name, type_val), value in zip_longest(self._type_hints(), args, fillvalue=MissingSentinel):
            if name in kwargs and value is not MissingSentinel:
                raise TypeError(f"got multiple values for argument {repr(name)}")

            if value is MissingSentinel and name in kwargs:
                value = kwargs[name]

            self._fields[name] = Field(
                name, type_val, value, getattr(self, f"parse_{name}", None), getattr(self, f"validate_{name}", None)
            )

    # noinspection PyMethodParameters
    @self_or_classmethod
    def _type_hints(self_or_cls) -> dict[str, Any]:
        fields = dict()
        if isinstance(self_or_cls, type):
            cls = self_or_cls
        else:
            cls = type(self_or_cls)
        for b in cls.__mro__[::-1]:
            base_fields = getattr(b, "_fields", None)
            if base_fields is not None:
                list_fields = {k: v for k, v in get_type_hints(b).items() if not k.startswith("_")}
                fields |= list_fields
        logger.debug(f"The obtained fields are: {fields!r}")
        fields |= super()._type_hints()
        yield from fields.items()


MinosModel = DeclarativeModel



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .abc import (
    DeclarativeModel,
    MinosModel,
)
from .aggregate import (
    Aggregate,
    AggregateDiff,
    AggregateRef,
)
from .networks import (
    Command,
    CommandReply,
    CommandStatus,
    Event,
)
from .value_object import (
    ValueObject,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from operator import (
    attrgetter,
)
from typing import (
    TYPE_CHECKING,
)
from uuid import (
    UUID,
)

from ...dynamic import (
    FieldsDiff,
)
from ..abc import (
    DeclarativeModel,
)

if TYPE_CHECKING:
    from .model import (
        Aggregate,
    )

logger = logging.getLogger(__name__)


class AggregateDiff(DeclarativeModel):
    """Aggregate Difference class."""

    uuid: UUID
    name: str
    version: int
    fields_diff: FieldsDiff

    @classmethod
    def from_difference(cls, a: Aggregate, b: Aggregate) -> AggregateDiff:
        """Build an ``AggregateDiff`` instance from the difference of two aggregates.

        :param a: One ``Aggregate`` instance.
        :param b: Another ``Aggregate`` instance.
        :return: An ``AggregateDiff`` instance.
        """
        logger.debug(f"Computing the {cls!r} between {a!r} and {b!r}...")

        if a.uuid != b.uuid:
            raise ValueError(
                f"To compute aggregate differences, both arguments must have same identifier. "
                f"Obtained: {a.uuid!r} vs {b.uuid!r}"
            )

        old, new = sorted([a, b], key=attrgetter("version"))

        fields_diff = FieldsDiff.from_difference(a, b, ignore=["uuid", "version"])

        return cls(new.uuid, new.classname, new.version, fields_diff)

    @classmethod
    def from_aggregate(cls, aggregate: Aggregate) -> AggregateDiff:
        """Build an ``AggregateDiff`` from an ``Aggregate`` (considering all fields as differences).

        :param aggregate: An ``Aggregate`` instance.
        :return: An ``AggregateDiff`` instance.
        """

        fields_diff = FieldsDiff.from_model(aggregate, ignore=["uuid", "version"])
        return cls(aggregate.uuid, aggregate.classname, aggregate.version, fields_diff)

    @classmethod
    def from_deleted_aggregate(cls, aggregate: Aggregate) -> AggregateDiff:
        """Build an ``AggregateDiff`` from an ``Aggregate`` (considering all fields as differences).

        :param aggregate: An ``Aggregate`` instance.
        :return: An ``AggregateDiff`` instance.
        """
        return cls(aggregate.uuid, aggregate.classname, aggregate.version, FieldsDiff.empty())

    @classmethod
    def simplify(cls, *args: AggregateDiff) -> AggregateDiff:
        """Simplify an iterable of aggregate differences into a single one.

        :param args: A sequence of ``FieldsDiff` instances.
        :return: An ``FieldsDiff`` instance.
        """
        args = sorted(args, key=attrgetter("version"))

        current = dict()
        for another in map(attrgetter("fields_diff"), args):
            # noinspection PyUnresolvedReferences
            current |= another.fields

        return cls(args[-1].uuid, args[-1].name, args[-1].version, FieldsDiff(current))



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from typing import (
    TYPE_CHECKING,
    Optional,
)

from .bucket import (
    BucketModel,
)

if TYPE_CHECKING:
    from ..abc import (
        Model,
    )
    from ..fields import (
        Field,
    )

logger = logging.getLogger(__name__)


class FieldsDiff(BucketModel):
    """Aggregate Difference class."""

    @classmethod
    def from_difference(cls, a: Model, b: Model, ignore: Optional[list[str]] = None) -> FieldsDiff:
        """Build an ``FieldsDiff`` instance from the difference of two aggregates.

        :param a: One ``Aggregate`` instance.
        :param b: Another ``Aggregate`` instance.
        :param ignore: List of fields to be ignored.
        :return: An ``FieldsDiff`` instance.
        """
        if ignore is None:
            ignore = list()

        logger.debug(f"Computing the {cls!r} between {a!r} and {b!r}...")
        fields = cls._diff(a.fields, b.fields)
        for name in ignore:
            fields.pop(name, None)

        return cls(fields)

    @staticmethod
    def _diff(a: dict[str, Field], b: dict[str, Field]) -> dict[str, Field]:
        """Compute the difference between ``a`` and ``b``.

        The implemented approach is equivalent to `dict(set(a.items()) -`set(b.items()))` but without requesting any
        hashing assumptions.

        :param a: The first dictionary of fields.
        :param b: The second dictionary of fields.
        :return: The differences dictionary between ``a`` and ``b``.
        """

        def _condition(key: str) -> bool:
            return key not in b or a[key] != b[key]

        return {key: a[key] for key in a if _condition(key)}

    @classmethod
    def from_model(cls, aggregate: Model, ignore: Optional[list[str]] = None) -> FieldsDiff:
        """Build an ``FieldsDiff`` from an ``Aggregate`` (considering all fields as differences).

        :param aggregate: An ``Aggregate`` instance.
        :param ignore: List of fields to be ignored.
        :return: An ``FieldsDiff`` instance.
        """
        if ignore is None:
            ignore = list()

        fields = dict(aggregate.fields)
        for name in ignore:
            fields.pop(name, None)

        return cls(fields)

    @classmethod
    def simplify(cls, *args: FieldsDiff) -> FieldsDiff:
        """Simplify an iterable of aggregate differences into a single one.

        :param args: A sequence of ``FieldsDiff` instances.
        :return: An ``FieldsDiff`` instance.
        """
        current = cls(args[0].fields)
        for another in args[1:]:
            current._fields |= another._fields
        return current



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from base64 import (
    b64decode,
    b64encode,
)
from typing import (
    Any,
    Generic,
    Iterable,
    NoReturn,
    Type,
    TypeVar,
    Union,
)

from ..exceptions import (
    EmptyMinosModelSequenceException,
    MinosModelException,
    MultiTypeMinosModelSequenceException,
)
from ..importlib import (
    classname,
)
from ..meta import (
    classproperty,
    property_or_classproperty,
    self_or_classmethod,
)
from ..protocol import (
    MinosAvroProtocol,
)
from .fields import (
    Field,
)
from .serializers import (
    AvroDataDecoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
)
from .types import (
    ModelType,
)

logger = logging.getLogger(__name__)

# def _process_aggregate(cls):
#     """
#     Get the list of the class arguments and define it as an AggregateField class
#     """
#     cls_annotations = cls.__dict__.get('__annotations__', {})
#     aggregate_fields = []
#     for name, type in cls_annotations.items():
#         attribute = getattr(cls, name, None)
#         aggregate_fields.append(
#             AggregateField(name=name, type=type, value=attribute)
#         )
#     setattr(cls, "_FIELDS", aggregate_fields)
#
#     # g get metaclass
#     meta_class = getattr(cls, "Meta", None)
#     if meta_class:
#         # meta class exist so get the information related
#         ...
#     return cls
#
#
# def aggregate(cls=None):
#     def wrap(cls):
#         return _process_aggregate(cls)
#
#     if cls is None:
#         return wrap
#
#     return wrap(cls)

T = TypeVar("T")


class Model(Generic[T]):
    """Base class for ``minos`` model entities."""

    _fields: dict[str, Field] = {}

    def __init__(self, fields: Union[Iterable[Field], dict[str, Field]] = None):
        """Class constructor.

        :param fields: Dictionary that contains the ``Field`` instances of the model indexed by name.
        """
        if fields is None:
            fields = dict()
        if not isinstance(fields, dict):
            fields = {field.name: field for field in fields}
        self._fields = fields

    @classmethod
    def from_avro_str(cls, raw: str, **kwargs) -> Union[T, list[T]]:
        """Build a single instance or a sequence of instances from bytes

        :param raw: A bytes data.
        :return: A single instance or a sequence of instances.
        """
        raw = b64decode(raw.encode())
        return cls.from_avro_bytes(raw, **kwargs)

    @classmethod
    def from_avro_bytes(cls, raw: bytes, **kwargs) -> Union[T, list[T]]:
        """Build a single instance or a sequence of instances from bytes

        :param raw: A bytes data.
        :return: A single instance or a sequence of instances.
        """
        schema = MinosAvroProtocol.decode_schema(raw)
        decoded = MinosAvroProtocol.decode(raw)

        # FIXME: Extend implementation of the `AvroDataDecoder` to avoid this fix.
        schema["name"] = "{}.fake.{}".format(*schema["name"].rsplit(".", 1))

        if isinstance(decoded, list):
            return [cls.from_avro(schema, d | kwargs) for d in decoded]
        return cls.from_avro(schema, decoded | kwargs)

    @classmethod
    def from_avro(cls, schema: Union[dict[str, Any], list[dict[str, Any]]], data: dict[str, Any]) -> T:
        """Build a new instance from the ``avro`` schema and data.

        :param schema: The avro schema of the model.
        :param data: The avro data of the model.
        :return: A new ``DynamicModel`` instance.
        """
        if isinstance(schema, list):
            schema = schema[-1]
        model_type: ModelType = AvroSchemaDecoder(schema).build()
        return AvroDataDecoder("", model_type).build(data)

    @classmethod
    def to_avro_str(cls, models: list[T]) -> str:
        """Create a bytes representation of the given object instances.

        :param models: A sequence of minos models.
        :return: A bytes object.
        """
        return b64encode(cls.to_avro_bytes(models)).decode()

    @classmethod
    def to_avro_bytes(cls, models: list[T]) -> bytes:
        """Create a bytes representation of the given object instances.

        :param models: A sequence of minos models.
        :return: A bytes object.
        """
        if len(models) == 0:
            raise EmptyMinosModelSequenceException("'models' parameter cannot be empty.")

        model_type = type(models[0])
        if not all(model_type == type(model) for model in models):
            raise MultiTypeMinosModelSequenceException(
                f"Every model must have type {model_type} to be valid. Found types: {[type(model) for model in models]}"
            )

        avro_schema = models[0].avro_schema
        # noinspection PyTypeChecker
        return MinosAvroProtocol().encode([model.avro_data for model in models], avro_schema)

    # noinspection PyMethodParameters
    @property_or_classproperty
    def model_type(self_or_cls) -> Type[T]:
        """Get the model type of the instance.

        :return: A ``ModelType`` instance.
        """
        # noinspection PyTypeChecker
        return ModelType.build(self_or_cls.classname, self_or_cls.type_hints)

    # noinspection PyMethodParameters
    @classproperty
    def classname(cls) -> str:
        """Compute the current class namespace.
        :return: An string object.
        """
        # noinspection PyTypeChecker
        return classname(cls)

    @property
    def fields(self) -> dict[str, Field]:
        """Fields getter"""
        return self._fields

    def __setattr__(self, key: str, value: Any) -> NoReturn:
        if self._fields is not None and key in self._fields:
            field_class: Field = self._fields[key]
            field_class.value = value
            self._fields[key] = field_class
        elif key.startswith("_"):
            super().__setattr__(key, value)
        else:
            raise MinosModelException(f"model attribute {key} doesn't exist")

    def __getattr__(self, item: str) -> Any:
        if self._fields is not None and item in self._fields:
            return self._fields[item].value
        else:
            raise AttributeError(f"{type(self).__name__!r} does not contain the {item!r} attribute.")

    # noinspection PyMethodParameters
    @property_or_classproperty
    def type_hints(self_or_cls) -> dict[str, type]:
        """Get the type hinting of the instance or class.

        :return: A dictionary in which the keys are the field names and the values are the types.
        """
        return dict(self_or_cls._type_hints())

    # noinspection PyMethodParameters
    @self_or_classmethod
    def _type_hints(self_or_cls) -> dict[str, Any]:
        if not isinstance(self_or_cls, type):
            yield from ((field.name, field.real_type) for field in self_or_cls.fields.values())

    # noinspection PyMethodParameters
    @property_or_classproperty
    def avro_schema(self_or_cls) -> list[dict[str, Any]]:
        """Compute the avro schema of the model.

        :return: A dictionary object.
        """
        # noinspection PyTypeChecker
        return [AvroSchemaEncoder("", self_or_cls.model_type).build()["type"]]

    @property
    def avro_data(self) -> dict[str, Any]:
        """Compute the avro data of the model.

        :return: A dictionary object.
        """
        return {name: field.avro_data for name, field in self.fields.items()}

    @property
    def avro_str(self) -> str:
        """Generate bytes representation of the current instance.

        :return: A bytes object.
        """
        # noinspection PyTypeChecker
        return b64encode(self.avro_bytes).decode()

    @property
    def avro_bytes(self) -> bytes:
        """Generate bytes representation of the current instance.

        :return: A bytes object.
        """
        # noinspection PyTypeChecker
        return MinosAvroProtocol().encode(self.avro_data, self.avro_schema)

    def __eq__(self, other: T) -> bool:
        return type(self) == type(other) and self.fields == other.fields

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable:
        # noinspection PyRedundantParentheses
        yield from self.fields.values()

    def __getitem__(self, item: str) -> Any:
        return getattr(self, item)

    def __setitem__(self, key: str, value: Any) -> NoReturn:
        setattr(self, key, value)

    def keys(self) -> Iterable[str]:
        """Get the field names.

        :return: An iterable of string values.
        """
        return self.fields.keys()

    def __repr__(self) -> str:
        fields_repr = ", ".join(repr(field) for field in self.fields.values())
        return f"{type(self).__name__}({fields_repr})"



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from typing import (
    Generic,
    Iterable,
    Optional,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from uuid import (
    UUID,
)

from .comparators import (
    is_aggregate_type,
    is_type_subclass,
)
from .model_refs import (
    ModelRef,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


def build_union(options: tuple[Type[T], ...]) -> Type[T]:
    """Build the union type base on the given options.

    :param options: A tuple of types.
    :return: The union of types.
    """
    if (
        len(options) == 2
        and is_aggregate_type(options[0])
        and is_type_subclass(options[1])
        and issubclass(options[1], UUID)
    ):
        return ModelRef[options[0]]

    return Union[options]


class TypeHintBuilder(Generic[T]):
    """Type Hint Builder class."""

    def __init__(self, value: T, base: Optional[Type[T]] = None):
        self._value = value
        self._base = base

    def build(self) -> Type[T]:
        """Build type hint from an instance..

        :return: A type.
        """
        return self._build(self._value, self._base)

    def _build(self, value: T, base: Optional[Type[T]]) -> Type[T]:
        if base is not None:
            if get_origin(base) is ModelRef:
                base = Union[(*get_args(base), UUID)]

            if get_origin(base) is Union:
                dynamic: Type[T] = self._build(value, None)
                options = tuple(
                    (dynamic if not len(get_args(static)) and issubclass(dynamic, static) else static)
                    for static in get_args(base)
                )
                return build_union(options)

        if isinstance(value, (tuple, list, set)):
            b1 = None if (base is None or len(get_args(base)) != 1) else get_args(base)[0]
            return type(value)[self._build_from_iterable(value, b1)]

        if isinstance(value, dict):
            b1, b2 = (None, None) if (base is None or len(get_args(base)) != 2) else get_args(base)
            return type(value)[
                self._build_from_iterable(value.keys(), b1), self._build_from_iterable(value.values(), b2)
            ]

        if hasattr(value, "model_type"):
            return value.model_type

        return type(value)

    def _build_from_iterable(self, values: Iterable[T], base: Optional[Type[T]]) -> Type[T]:
        values = tuple(values)
        if len(values) == 0:
            return base

        options = tuple(self._build(value, base) for value in values)
        return build_union(options)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from typing import (
    TYPE_CHECKING,
    AsyncIterator,
    NoReturn,
)

from ...configuration import (
    MinosConfig,
)
from ...exceptions import (
    MinosSnapshotAggregateNotFoundException,
    MinosSnapshotDeletedAggregateException,
)
from ..abc import (
    MinosSnapshot,
)
from ..entries import (
    SnapshotEntry,
)
from .abc import (
    PostgreSqlSnapshotSetup,
)
from .builders import (
    PostgreSqlSnapshotBuilder,
)

if TYPE_CHECKING:

    from ...model import (
        Aggregate,
    )


class PostgreSqlSnapshot(PostgreSqlSnapshotSetup, MinosSnapshot):
    """Minos Snapshot Reader class."""

    builder: PostgreSqlSnapshotBuilder

    def __init__(self, *args, builder: PostgreSqlSnapshotBuilder, **kwargs):
        super().__init__(*args, **kwargs)
        self.builder = builder

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> PostgreSqlSnapshot:
        builder = PostgreSqlSnapshotBuilder.from_config(*args, config=config, **kwargs)
        return cls(*args, builder=builder, **config.snapshot._asdict(), **kwargs)

    async def _setup(self) -> NoReturn:
        await self.builder.setup()
        await super()._setup()

    async def _destroy(self) -> NoReturn:
        await super()._destroy()
        await self.builder.destroy()

    async def get(self, aggregate_name: str, ids: list[int], **kwargs) -> AsyncIterator[Aggregate]:
        """Retrieve an asynchronous iterator that provides the requested ``Aggregate`` instances.

        :param aggregate_name: Class name of the ``Aggregate`` to be retrieved.
        :param ids: List of identifiers to be retrieved.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that provides the requested ``Aggregate`` instances.
        """
        # noinspection PyShadowingBuiltins
        if not await self.builder.are_synced(aggregate_name, ids):
            await self.builder.dispatch()

        async for item in self._get(aggregate_name, ids):
            yield item.aggregate

    async def _get(self, aggregate_name: str, ids: list[int]) -> AsyncIterator[SnapshotEntry]:
        ids = tuple(set(ids))
        parameters = (aggregate_name, ids)

        with (await self.cursor()) as cursor:
            async with cursor.begin():

                await cursor.execute(_CHECK_MULTIPLE_ENTRIES_QUERY, parameters)
                total_count, not_null_count = await cursor.fetchone()

                await cursor.execute(_SELECT_MULTIPLE_ENTRIES_QUERY, parameters)

                if total_count != len(ids):
                    # noinspection PyUnresolvedReferences
                    found = {row[0] async for row in cursor}
                    missing = set(ids) - found
                    raise MinosSnapshotAggregateNotFoundException(f"Some aggregates could not be found: {missing!r}")

                if not_null_count != len(ids):
                    # noinspection PyUnresolvedReferences
                    found = {row[0] async for row in cursor if row[3]}
                    missing = set(ids) - found
                    raise MinosSnapshotDeletedAggregateException(f"Some aggregates are already deleted: {missing!r}")

                async for row in cursor:
                    # noinspection PyArgumentList
                    yield SnapshotEntry(*row)

    # noinspection PyUnusedLocal
    async def select(self, *args, **kwargs) -> AsyncIterator[SnapshotEntry]:
        """Select a sequence of ``MinosSnapshotEntry`` objects.

        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A sequence of ``MinosSnapshotEntry`` objects.
        """
        async for row in self.submit_query_and_iter(_SELECT_ALL_ENTRIES_QUERY):
            yield SnapshotEntry(*row)


_SELECT_ALL_ENTRIES_QUERY = """
SELECT aggregate_uuid, aggregate_name, version, data, created_at, updated_at
FROM snapshot
ORDER BY updated_at;
""".strip()

_SELECT_MULTIPLE_ENTRIES_QUERY = """
SELECT aggregate_uuid, aggregate_name, version, data, created_at, updated_at
FROM snapshot
WHERE aggregate_name = %s AND aggregate_uuid IN %s
ORDER BY updated_at;
""".strip()

_CHECK_MULTIPLE_ENTRIES_QUERY = """
SELECT COUNT(*) as total_count, COUNT(data) as not_null_count
FROM snapshot
WHERE aggregate_name = %s AND aggregate_uuid IN %s;
""".strip()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Mapping,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from uuid import (
    UUID,
)

from ...exceptions import (
    MinosMalformedAttributeException,
    MinosReqAttributeException,
    MinosTypeAttributeException,
)
from ..types import (
    MissingSentinel,
    ModelRef,
    ModelType,
    NoneType,
    TypeHintBuilder,
    is_aggregate_type,
    is_model_subclass,
    is_type_subclass,
)

if TYPE_CHECKING:
    from ..fields import (
        Field,
    )
logger = logging.getLogger(__name__)

T = TypeVar("T")


class AvroDataDecoder:
    """Avro Data Decoder class."""

    def __init__(self, field_name: str, field_type: Type):
        self._name = field_name
        self._type = field_type

    @classmethod
    def from_field(cls, field: Field) -> AvroDataDecoder:
        """Build a new instance from a ``Field``.

        :param field: The model field.
        :return: A new avro schema builder instance.
        """
        return cls(field.name, field.type)

    def build(self, data: Any) -> Any:
        """Cast data type according to the field definition..

        :param data: Data to be casted.
        :return: Casted object.
        """
        return self._cast_value(self._type, data)

    def _cast_value(self, type_field: Type, data: Any) -> Any:
        if type_field is Any:
            type_field = TypeHintBuilder(data).build()
        origin = get_origin(type_field)
        if origin is not Union:
            return self._cast_single_value(type_field, data)
        return self._cast_union_value(type_field, data)

    def _cast_union_value(self, type_field: Type, data: Any) -> Any:
        alternatives = get_args(type_field)
        for alternative_type in alternatives:
            try:
                return self._cast_single_value(alternative_type, data)
            except (MinosTypeAttributeException, MinosReqAttributeException):
                pass

        if type_field is not NoneType:
            if data is None:
                raise MinosReqAttributeException(f"{self._name!r} field is {None!r}.")

            if data is MissingSentinel:
                raise MinosReqAttributeException(f"{self._name!r} field is missing.")

        raise MinosTypeAttributeException(self._name, type_field, data)

    def _cast_single_value(self, type_field: Type, data: Any) -> Any:
        if type_field is NoneType:
            return self._cast_none_value(type_field, data)

        if data is None:
            raise MinosReqAttributeException(f"{self._name!r} field is '{None!r}'.")

        if data is MissingSentinel:
            raise MinosReqAttributeException(f"{self._name!r} field is missing.")

        if is_type_subclass(type_field):
            if issubclass(type_field, bool):
                return self._cast_bool(data)

            if issubclass(type_field, int):
                return self._cast_int(type_field, data)

            if issubclass(type_field, float):
                return self._cast_float(data)

            if issubclass(type_field, str):
                return self._cast_string(data)

            if issubclass(type_field, bytes):
                return self._cast_bytes(data)

            if issubclass(type_field, datetime):
                return self._cast_datetime(data)

            if issubclass(type_field, date):
                return self._cast_date(data)

            if issubclass(type_field, time):
                return self._cast_time(data)

            if issubclass(type_field, UUID):
                return self._cast_uuid(data)

            if isinstance(type_field, ModelType):
                return self._cast_model_type(type_field, data)

            if is_model_subclass(type_field):
                return self._cast_model(type_field, data)

        return self._cast_composed_value(type_field, data)

    def _cast_none_value(self, type_field: Type, data: Any) -> Any:
        if data is None or data is MissingSentinel:
            return None

        raise MinosTypeAttributeException(self._name, type_field, data)

    def _cast_int(self, type_field, data: Any) -> int:
        try:
            return type_field(data)
        except (ValueError, TypeError):
            raise MinosTypeAttributeException(self._name, type_field, data)

    def _cast_float(self, data: Any) -> float:
        try:
            return float(data)
        except (ValueError, TypeError):
            raise MinosTypeAttributeException(self._name, float, data)

    def _cast_bool(self, data: Any) -> bool:
        if not isinstance(data, bool):
            raise MinosTypeAttributeException(self._name, bool, data)
        return data

    def _cast_string(self, data: Any) -> str:
        if not isinstance(data, str):
            raise MinosTypeAttributeException(self._name, str, data)
        return data

    def _cast_bytes(self, data: Any) -> bytes:
        if not isinstance(data, bytes):
            raise MinosTypeAttributeException(self._name, bytes, data)
        return data

    def _cast_date(self, data: Any) -> date:
        if isinstance(data, date):
            return data
        elif isinstance(data, int):
            return date(1970, 1, 1) + timedelta(days=data)
        raise MinosTypeAttributeException(self._name, date, data)

    def _cast_time(self, data: Any) -> time:
        if isinstance(data, time):
            return data
        if isinstance(data, int):
            return (datetime(1, 1, 1) + timedelta(microseconds=data)).time()
        raise MinosTypeAttributeException(self._name, time, data)

    def _cast_datetime(self, data: Any) -> datetime:
        if isinstance(data, datetime):
            return data
        if isinstance(data, int):
            return datetime(1970, 1, 1) + data * timedelta(microseconds=1)
        raise MinosTypeAttributeException(self._name, datetime, data)

    def _cast_uuid(self, data: Any) -> UUID:
        if isinstance(data, UUID):
            return data
        elif isinstance(data, str):
            try:
                return UUID(hex=data)
            except ValueError:
                pass
        elif isinstance(data, bytes):
            try:
                return UUID(bytes=data)
            except ValueError:
                pass
        raise MinosTypeAttributeException(self._name, UUID, data)

    def _cast_model_type(self, type_field: ModelType, data: Any) -> Any:
        if isinstance(data, dict):
            data = {k: self._cast_value(v, data[k]) for k, v in type_field.type_hints.items()}
            return type_field(**data)

        if hasattr(data, "model_type") and type_field == data.model_type:
            return data

        raise MinosTypeAttributeException(self._name, type_field, data)

    def _cast_model(self, type_field: Type, data: Any) -> Any:
        if not isinstance(data, type_field):
            raise MinosTypeAttributeException(self._name, type_field, data)
        return data

    def _cast_composed_value(self, type_field: Type, data: Any) -> Any:
        origin_type = get_origin(type_field)
        if origin_type is None:
            raise MinosMalformedAttributeException(f"{self._name!r} field is malformed. Type: '{type_field}'.")

        if origin_type is list:
            return self._convert_list(data, type_field)

        if origin_type is dict:
            return self._convert_dict(data, type_field)

        if origin_type is ModelRef:
            return self._convert_model_ref(data, type_field)

        raise MinosTypeAttributeException(self._name, type_field, data)

    def _convert_list(self, data: list, type_values: Any) -> list[Any]:
        type_values = get_args(type_values)[0]
        if not isinstance(data, list):
            raise MinosTypeAttributeException(self._name, list, data)

        return self._convert_list_params(data, type_values)

    def _convert_dict(self, data: list, type_field: Type) -> dict[str, Any]:
        type_keys, type_values = get_args(type_field)
        if not isinstance(data, dict):
            raise MinosTypeAttributeException(self._name, dict, data)

        if type_keys is not str:
            raise MinosMalformedAttributeException(f"dictionary keys must be {str!r}. Obtained: {type_keys!r}")

        return self._convert_dict_params(data, type_keys, type_values)

    def _convert_dict_params(self, data: Mapping, type_keys: Type, type_values: Type) -> dict[Any, Any]:
        keys = self._convert_list_params(data.keys(), type_keys)
        values = self._convert_list_params(data.values(), type_values)
        return dict(zip(keys, values))

    def _convert_model_ref(self, data: Any, type_field: Type) -> Any:
        inner_type = get_args(type_field)[0]
        if not is_aggregate_type(inner_type):
            raise MinosMalformedAttributeException(
                f"'ModelRef[T]' T type must follow the 'Aggregate' protocol. Obtained: {inner_type!r}"
            )

        return self._cast_value(Union[inner_type, UUID], data)

    def _convert_list_params(self, data: Iterable, type_params: Type) -> list[Any]:
        """
        check if the parameters list are equal to @type_params type
        """
        converted = list()
        for item in data:
            value = self._cast_value(type_params, item)
            converted.append(value)
        return converted



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from datetime import (
    datetime,
)
from enum import (
    Enum,
)
from typing import (
    TYPE_CHECKING,
    Iterable,
    Optional,
    Type,
    Union,
)
from uuid import (
    UUID,
)

from ..exceptions import (
    MinosRepositoryUnknownActionException,
)
from ..importlib import (
    import_module,
)

if TYPE_CHECKING:
    from ..model import (
        Aggregate,
        AggregateDiff,
    )


class RepositoryAction(Enum):
    """Enum class that describes the available repository actions."""

    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"

    @classmethod
    def value_of(cls, value: str) -> Optional[RepositoryAction]:
        """Get the action based on its text representation."""
        for item in cls.__members__.values():
            if item.value == value:
                return item
        raise MinosRepositoryUnknownActionException(
            f"The given value does not match with any enum items. Obtained {value}"
        )


class RepositoryEntry:
    """Class that represents an entry (or row) on the events repository database which stores the aggregate changes."""

    __slots__ = "aggregate_uuid", "aggregate_name", "version", "data", "id", "action", "created_at"

    # noinspection PyShadowingBuiltins
    def __init__(
        self,
        aggregate_uuid: UUID,
        aggregate_name: str,
        version: int,
        data: Union[bytes, memoryview] = bytes(),
        id: Optional[int] = None,
        action: Optional[Union[str, RepositoryAction]] = None,
        created_at: Optional[datetime] = None,
    ):
        if isinstance(data, memoryview):
            data = data.tobytes()
        if action is not None and isinstance(action, str):
            action = RepositoryAction.value_of(action)

        self.aggregate_uuid = aggregate_uuid
        self.aggregate_name = aggregate_name
        self.version = version
        self.data = data

        self.id = id
        self.action = action
        self.created_at = created_at

    @classmethod
    def from_aggregate_diff(cls, aggregate_diff: AggregateDiff) -> RepositoryEntry:
        """Build a new instance from an ``Aggregate``.

        :param aggregate_diff: The aggregate instance.
        :return: A new ``RepositoryEntry`` instance.
        """
        # noinspection PyTypeChecker
        return cls(
            aggregate_diff.uuid, aggregate_diff.name, aggregate_diff.version, aggregate_diff.fields_diff.avro_bytes
        )

    @property
    def aggregate_cls(self) -> Type[Aggregate]:
        """Load the concrete ``Aggregate`` class.

        :return: A ``Type`` object.
        """
        # noinspection PyTypeChecker
        return import_module(self.aggregate_name)

    @property
    def aggregate_diff(self) -> AggregateDiff:
        """Get the stored ``AggregateDiff`` instance.

        :return: An ``AggregateDiff`` instance.
        """
        from ..model import (
            AggregateDiff,
            FieldsDiff,
        )

        return AggregateDiff(
            self.aggregate_uuid, self.aggregate_name, self.version, FieldsDiff.from_avro_bytes(self.data)
        )

    def __eq__(self, other: "RepositoryEntry") -> bool:
        return type(self) == type(other) and tuple(self) == tuple(other)

    def __hash__(self) -> int:
        return hash(tuple(self))

    def __iter__(self) -> Iterable:
        yield from (
            self.aggregate_uuid,
            self.aggregate_name,
            self.version,
            self.data,
            self.id,
            self.action,
            self.created_at,
        )

    def __repr__(self):
        return (
            f"{type(self).__name__}("
            f"aggregate_uuid={self.aggregate_uuid!r}, aggregate_name={self.aggregate_name!r}, "
            f"version={self.version!r}, data={self.data!r}, "
            f"id={self.id!r}, action={self.action!r}, created_at={self.created_at!r})"
        )



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from typing import (
    Any,
)

from ...exceptions import (
    MinosImmutableClassException,
)
from .abc import (
    DeclarativeModel,
)


class ValueObject(DeclarativeModel):
    """Value Object class."""

    def __setattr__(self, key: str, value: Any):
        if key.startswith("_"):
            super().__setattr__(key, value)
        else:
            raise MinosImmutableClassException("modification of an immutable value object not allowed")



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
    Generic,
    Iterable,
    TypeVar,
    Union,
)

from ..abc import (
    Model,
)
from ..fields import (
    Field,
)
from ..types import (
    ModelType,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


class DynamicModel(Model, Generic[T]):
    """Base class for ``minos`` dynamic model entities"""

    def __init__(self, fields: Union[Iterable[Field], dict[str, Field]], **kwargs):
        super().__init__(fields)

    @classmethod
    def from_model_type(cls, model_type: ModelType, data: dict[str, Any]) -> T:
        """Build a ``DynamicModel`` from a ``ModelType`` and ``data``.

        :param model_type: ``ModelType`` object containing the DTO's structure
        :param data: A dictionary containing the values to be stored on the DTO.
        :return: A new ``DynamicModel`` instance.
        """
        return cls(fields={k: Field(k, v, data[k]) for k, v in model_type.type_hints.items()})



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .abc import (
    PostgreSqlSnapshotSetup,
)
from .builders import (
    PostgreSqlSnapshotBuilder,
)
from .snapshot import (
    PostgreSqlSnapshot,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

__version__ = "0.1.4"

from .configuration import (
    BROKER,
    COMMANDS,
    CONTROLLER,
    ENDPOINT,
    EVENTS,
    QUEUE,
    REPOSITORY,
    REST,
    SAGA,
    SERVICE,
    SNAPSHOT,
    STORAGE,
    MinosConfig,
    MinosConfigAbstract,
)
from .constants import (
    NULL_UUID,
)
from .database import (
    PostgreSqlMinosDatabase,
)
from .exceptions import (
    EmptyMinosModelSequenceException,
    MinosAttributeValidationException,
    MinosBrokerException,
    MinosBrokerNotProvidedException,
    MinosConfigException,
    MinosException,
    MinosImmutableClassException,
    MinosImportException,
    MinosMalformedAttributeException,
    MinosMessageException,
    MinosModelAttributeException,
    MinosModelException,
    MinosParseAttributeException,
    MinosPreviousVersionSnapshotException,
    MinosProtocolException,
    MinosRepositoryException,
    MinosRepositoryManuallySetAggregateIdentifierException,
    MinosRepositoryManuallySetAggregateVersionException,
    MinosRepositoryNotProvidedException,
    MinosRepositoryUnknownActionException,
    MinosReqAttributeException,
    MinosSagaManagerException,
    MinosSagaManagerNotProvidedException,
    MinosSnapshotAggregateNotFoundException,
    MinosSnapshotDeletedAggregateException,
    MinosSnapshotException,
    MinosSnapshotNotProvidedException,
    MinosTypeAttributeException,
    MultiTypeMinosModelSequenceException,
)
from .importlib import (
    classname,
    import_module,
)
from .injectors import (
    DependencyInjector,
)
from .launchers import (
    EntrypointLauncher,
)
from .messages import (
    Request,
    Response,
    ResponseException,
)
from .meta import (
    classproperty,
    property_or_classproperty,
    self_or_classmethod,
)
from .model import (
    Aggregate,
    AggregateDiff,
    AggregateRef,
    AvroDataDecoder,
    AvroDataEncoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
    BucketModel,
    Command,
    CommandReply,
    CommandStatus,
    DataTransferObject,
    Decimal,
    DeclarativeModel,
    DynamicModel,
    Enum,
    Event,
    Field,
    FieldsDiff,
    Fixed,
    MinosModel,
    MissingSentinel,
    Model,
    ModelField,
    ModelRef,
    ModelRefExtractor,
    ModelRefInjector,
    ModelType,
    NoneType,
    TypeHintBuilder,
    TypeHintComparator,
    ValueObject,
)
from .networks import (
    MinosBroker,
)
from .protocol import (
    MinosAvroDatabaseProtocol,
    MinosAvroMessageProtocol,
    MinosAvroProtocol,
    MinosBinaryProtocol,
    MinosJsonBinaryProtocol,
)
from .repository import (
    InMemoryRepository,
    MinosRepository,
    PostgreSqlRepository,
    RepositoryAction,
    RepositoryEntry,
)
from .saga import (
    MinosSagaManager,
)
from .setup import (
    MinosSetup,
)
from .snapshot import (
    InMemorySnapshot,
    MinosSnapshot,
    PostgreSqlSnapshot,
    PostgreSqlSnapshotBuilder,
    PostgreSqlSnapshotSetup,
    SnapshotEntry,
)
from .storage import (
    MinosStorage,
    MinosStorageLmdb,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from typing import (
    TYPE_CHECKING,
    AsyncIterator,
)
from uuid import (
    UUID,
)

from ..exceptions import (
    MinosSnapshotAggregateNotFoundException,
    MinosSnapshotDeletedAggregateException,
)
from ..repository import (
    MinosRepository,
    RepositoryAction,
)
from .abc import (
    MinosSnapshot,
)

if TYPE_CHECKING:
    from ..model import (
        Aggregate,
    )


class InMemorySnapshot(MinosSnapshot):
    """In Memory Snapshot class."""

    async def get(self, aggregate_name: str, uuids: list[UUID], **kwargs) -> AsyncIterator[Aggregate]:
        """Retrieve an asynchronous iterator that provides the requested ``Aggregate`` instances.

        :param aggregate_name: Class name of the ``Aggregate`` to be retrieved.
        :param uuids: List of identifiers to be retrieved.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that provides the requested ``Aggregate`` instances.
        """
        iterable = map(lambda uuid: self._get_one(aggregate_name, uuid, **kwargs), uuids)

        for item in iterable:
            yield await item

    # noinspection PyShadowingBuiltins
    @staticmethod
    async def _get_one(aggregate_name: str, uuid: UUID, _repository: MinosRepository, **kwargs) -> Aggregate:
        from operator import (
            attrgetter,
        )

        # noinspection PyTypeChecker
        entries = [v async for v in _repository.select(aggregate_name=aggregate_name, aggregate_uuid=uuid)]
        if not len(entries):
            raise MinosSnapshotAggregateNotFoundException(f"Not found any entries for the {uuid!r} id.")

        entries.sort(key=attrgetter("version"))
        if entries[-1].action == RepositoryAction.DELETE:
            raise MinosSnapshotDeletedAggregateException(f"The {uuid!r} id points to an already deleted aggregate.")

        cls = entries[0].aggregate_cls
        instance: Aggregate = cls.from_diff(entries[0].aggregate_diff, _repository=_repository, **kwargs)
        for entry in entries[1:]:
            instance.apply_diff(entry.aggregate_diff)

        return instance



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from typing import (
    Generic,
    TypeVar,
)

from .abc import (
    DynamicModel,
)

T = TypeVar("T")


class BucketModel(DynamicModel, Generic[T]):
    """Bucket Model class."""

    @classmethod
    def empty(cls) -> T:
        """Build an empty ``BucketModel`` instance.

        :return: A ``BucketModel`` instance.
        """
        return cls(dict())



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    TYPE_CHECKING,
    AsyncIterator,
)

from ..setup import (
    MinosSetup,
)

if TYPE_CHECKING:
    from ..model import (
        Aggregate,
    )


class MinosSnapshot(ABC, MinosSetup):
    """Base Snapshot class."""

    @abstractmethod
    async def get(self, aggregate_name: str, ids: list[int], **kwargs) -> AsyncIterator[Aggregate]:
        """Retrieve an asynchronous iterator that provides the requested ``Aggregate`` instances.

        :param aggregate_name: Class name of the ``Aggregate`` to be retrieved.
        :param ids: List of identifiers to be retrieved.
        :param kwargs: Additional named arguments.
        :return: An asynchronous iterator that provides the requested ``Aggregate`` instances.
        """



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from uuid import (
    UUID,
)

NULL_UUID = UUID("00000000-0000-0000-0000-000000000000")



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .diff import (
    AggregateDiff,
)
from .model import (
    Aggregate,
)
from .ref import (
    AggregateRef,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    TYPE_CHECKING,
    Any,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from uuid import (
    UUID,
)

from ..types import (
    ModelRef,
    ModelType,
    NoneType,
    is_model_subclass,
    is_type_subclass,
)
from .constants import (
    AVRO_ARRAY,
    AVRO_BOOLEAN,
    AVRO_BYTES,
    AVRO_DATE,
    AVRO_DOUBLE,
    AVRO_INT,
    AVRO_MAP,
    AVRO_NULL,
    AVRO_STRING,
    AVRO_TIME,
    AVRO_TIMESTAMP,
    AVRO_UUID,
)

if TYPE_CHECKING:
    from ..fields import Field  # pragma: no cover

logger = logging.getLogger(__name__)

T = TypeVar("T")


class AvroSchemaEncoder:
    """Avro Schema Encoder class."""

    def __init__(self, field_name: str, field_type: Type):
        self._name = field_name
        self._type = field_type

    @classmethod
    def from_field(cls, field: Field) -> AvroSchemaEncoder:
        """Build a new instance from a ``Field``.

        :param field: The model field.
        :return: A new avro schema builder instance.
        """
        return cls(field.name, field.real_type)

    def build(self) -> dict[str, Any]:
        """Build the avro schema for the given field.

        :return: A dictionary object.
        """

        return {"name": self._name, "type": self._build_schema(self._type)}

    def _build_schema(self, type_field: Type) -> Any:
        origin = get_origin(type_field)
        if origin is not Union:
            return self._build_single_schema(type_field)
        return self._build_union_schema(type_field)

    def _build_union_schema(self, type_field: Type) -> Any:
        ans = list()
        alternatives = get_args(type_field)
        for alternative_type in alternatives:
            step = self._build_single_schema(alternative_type)
            if isinstance(step, list):
                ans += step
            else:
                ans.append(step)
        return ans

    def _build_single_schema(self, type_field: Type) -> Any:
        if type_field is Any:
            # FIXME: This is a design decision that must be revisited in the future.
            return AVRO_NULL

        if is_type_subclass(type_field):
            if issubclass(type_field, NoneType):
                return AVRO_NULL

            if issubclass(type_field, bool):
                return AVRO_BOOLEAN

            if issubclass(type_field, int):
                return AVRO_INT

            if issubclass(type_field, float):
                return AVRO_DOUBLE

            if issubclass(type_field, str):
                return AVRO_STRING

            if issubclass(type_field, bytes):
                return AVRO_BYTES

            if issubclass(type_field, datetime):
                return AVRO_TIMESTAMP

            if issubclass(type_field, date):
                return AVRO_DATE

            if issubclass(type_field, time):
                return AVRO_TIME

            if issubclass(type_field, UUID):
                return AVRO_UUID

            if isinstance(type_field, ModelType):
                return self._build_model_type_schema(type_field)

            if is_model_subclass(type_field):
                return self._build_model_schema(type_field)

        return self._build_composed_schema(type_field)

    def _build_model_type_schema(self, type_field: ModelType) -> Any:
        namespace = type_field.namespace
        if len(namespace) > 0 and len(self._name) > 0:
            namespace = f"{type_field.namespace}.{self._name}"
        schema = {
            "name": type_field.name,
            "namespace": namespace,
            "type": "record",
            "fields": [AvroSchemaEncoder(k, v).build() for k, v in type_field.type_hints.items()],
        }
        return schema

    def _build_model_schema(self, type_field: Type) -> Any:
        def _patch_namespace(s: dict) -> dict:
            if len(self._name) > 0:
                s["namespace"] += f".{self._name}"
            return s

        # noinspection PyUnresolvedReferences
        return [_patch_namespace(s) for s in type_field.avro_schema]

    def _build_composed_schema(self, type_field: Type) -> Any:
        origin_type = get_origin(type_field)

        if origin_type is list:
            return self._build_list_schema(type_field)

        if origin_type is dict:
            return self._build_dict_schema(type_field)

        if origin_type is ModelRef:
            return self._build_model_ref_schema(type_field)

        raise ValueError(f"Given field type is not supported: {type_field}")  # pragma: no cover

    def _build_list_schema(self, type_field: Type) -> dict[str, Any]:
        return {"type": AVRO_ARRAY, "items": self._build_schema(get_args(type_field)[0])}

    def _build_dict_schema(self, type_field: Type) -> dict[str, Any]:
        return {"type": AVRO_MAP, "values": self._build_schema(get_args(type_field)[1])}

    def _build_model_ref_schema(self, type_field: Type) -> Union[bool, Any]:
        return self._build_schema(Union[get_args(type_field)[0], UUID])



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""

from __future__ import (
    annotations,
)

from enum import (
    IntEnum,
)
from typing import (
    Any,
)
from uuid import (
    UUID,
)

from .abc import (
    DeclarativeModel,
)
from .aggregate import (
    AggregateDiff,
)


class Command(DeclarativeModel):
    """Base Command class."""

    topic: str
    data: Any
    saga: UUID
    reply_topic: str


class CommandReply(DeclarativeModel):
    """Base Command class."""

    topic: str
    data: Any
    saga: UUID
    status: CommandStatus


class CommandStatus(IntEnum):
    """Command Status class."""

    SUCCESS = 200
    ERROR = 400
    SYSTEM_ERROR = 500


class Event(DeclarativeModel):
    """Base Event class."""

    topic: str
    data: AggregateDiff



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from datetime import (
    date,
    datetime,
    time,
    timedelta,
)
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
)
from uuid import (
    UUID,
)

from ...exceptions import (
    MinosMalformedAttributeException,
)

if TYPE_CHECKING:
    from ..fields import Field  # pragma: no cover
logger = logging.getLogger(__name__)

T = TypeVar("T")


class AvroDataEncoder:
    """Avro Data Encoder class."""

    def __init__(self, value: Any):
        self._value = value

    @classmethod
    def from_field(cls, field: Field) -> AvroDataEncoder:
        """Build a new instance from a ``Field``.

        :param field: The model field.
        :return: A new avro schema builder instance.
        """
        return cls(field.value)

    def build(self):
        """Build a avro data representation based on the content of the given field.

        :return: A `avro`-compatible data.
        """
        return self._to_avro_raw(self._value)

    def _to_avro_raw(self, value: Any) -> Any:
        if value is None:
            return None

        if isinstance(value, (str, int, bool, float, bytes)):
            return value

        if isinstance(value, datetime):
            return self._datetime_to_avro_raw(value)

        if isinstance(value, date):
            return self._date_to_avro_raw(value)

        if isinstance(value, time):
            return self._time_to_avro_raw(value)

        if isinstance(value, UUID):
            return self._uuid_to_avro_raw(value)

        if isinstance(value, list):
            return [self._to_avro_raw(v) for v in value]

        if isinstance(value, dict):
            return {k: self._to_avro_raw(v) for k, v in value.items()}

        if hasattr(value, "avro_data"):
            return value.avro_data

        raise MinosMalformedAttributeException(f"Given type is not supported: {type(value)!r} ({value!r})")

    @staticmethod
    def _date_to_avro_raw(value: date) -> int:
        return (value - date(1970, 1, 1)).days

    @staticmethod
    def _time_to_avro_raw(value: time) -> int:
        return (datetime.combine(date(1, 1, 1), value) - datetime(1, 1, 1)) // timedelta(microseconds=1)

    @staticmethod
    def _datetime_to_avro_raw(value: datetime) -> int:
        return (value - datetime(1970, 1, 1, tzinfo=value.tzinfo)) // timedelta(microseconds=1)

    @staticmethod
    def _uuid_to_avro_raw(value: UUID) -> str:
        return str(value)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .avro_data_decoder import (
    AvroDataDecoder,
)
from .avro_data_encoder import (
    AvroDataEncoder,
)
from .avro_schema_decoder import (
    AvroSchemaDecoder,
)
from .avro_schema_encoder import (
    AvroSchemaEncoder,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from datetime import (
    date,
    datetime,
    time,
)
from typing import (
    Any,
    Optional,
    Type,
    TypeVar,
    Union,
)
from uuid import (
    UUID,
)

from ...exceptions import (
    MinosMalformedAttributeException,
)
from ..types import (
    ModelType,
    NoneType,
    build_union,
)
from .constants import (
    AVRO_ARRAY,
    AVRO_BOOLEAN,
    AVRO_BYTES,
    AVRO_DATE,
    AVRO_DOUBLE,
    AVRO_FLOAT,
    AVRO_INT,
    AVRO_MAP,
    AVRO_NULL,
    AVRO_RECORD,
    AVRO_STRING,
    AVRO_TIME,
    AVRO_TIMESTAMP,
    AVRO_UUID,
)

logger = logging.getLogger(__name__)

T = TypeVar("T")


class AvroSchemaDecoder:
    """Avro Schema Decoder class."""

    def __init__(self, schema: dict):
        self._schema = schema

    def build(self) -> Type[T]:
        """Build type from given avro schema item.

        :return: A dictionary object.
        """
        built_type = self._build_type(self._schema)
        return built_type

    def _build_type(self, schema: Union[dict, list, str]) -> Type[T]:
        if isinstance(schema, dict):
            return self._build_type_from_dict(schema)
        elif isinstance(schema, list):
            return self._build_type_from_list(schema)
        else:
            return self._build_simple_type(schema)

    def _build_type_from_list(self, schema: list[Any]) -> Type[T]:
        options = tuple(self._build_type(entry) for entry in schema)
        return build_union(options)

    def _build_type_from_dict(self, schema: dict) -> Type[T]:
        if "logicalType" in schema:
            return self._build_logical_type(schema["logicalType"])
        elif schema["type"] == AVRO_ARRAY:
            return self._build_list_type(schema["items"])
        elif schema["type"] == AVRO_MAP:
            return self._build_dict_type(schema["values"])
        elif schema["type"] == AVRO_RECORD:
            return self._build_record_type(schema["name"], schema.get("namespace", None), schema["fields"])
        else:
            return self._build_type(schema["type"])

    @staticmethod
    def _build_logical_type(type_field: str) -> Type[T]:
        if type_field == AVRO_DATE["logicalType"]:
            return date
        if type_field == AVRO_TIME["logicalType"]:
            return time
        if type_field == AVRO_TIMESTAMP["logicalType"]:
            return datetime
        if type_field == AVRO_UUID["logicalType"]:
            return UUID
        raise MinosMalformedAttributeException(f"Given logical field type is not supported: {type_field!r}")

    def _build_list_type(self, items: Union[dict, str, Any] = None) -> Type[T]:
        return list[self._build_type(items)]

    def _build_dict_type(self, values: Union[dict, str, Any] = None) -> Type[T]:
        return dict[str, self._build_type(values)]

    def _build_record_type(self, name: str, namespace: Optional[str], fields: list[dict[str, Any]]) -> Type[T]:
        def _unpatch_namespace(mt: Type[T]) -> Type[T]:
            try:
                mt.namespace, _ = mt.namespace.rsplit(".", 1)
            except ValueError:
                pass
            return mt

        model_type = ModelType.build(
            name, {field["name"]: self._build_type(field["type"]) for field in fields}, namespace
        )

        model_type = _unpatch_namespace(model_type)

        return model_type

    @staticmethod
    def _build_simple_type(type_field: str) -> Type[T]:
        if type_field == AVRO_NULL:
            return NoneType
        if type_field == AVRO_INT:
            return int
        if type_field == AVRO_BOOLEAN:
            return bool
        if type_field == AVRO_FLOAT:
            return float
        if type_field == AVRO_DOUBLE:
            return float
        if type_field == AVRO_STRING:
            return str
        if type_field == AVRO_BYTES:
            return bytes

        raise MinosMalformedAttributeException(f"Given field type is not supported: {type_field!r}")



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from typing import (
    Any,
    Generic,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
from uuid import (
    UUID,
)

from .model_refs import (
    ModelRef,
)
from .model_types import (
    ModelType,
)


def is_model_subclass(type_field: Any) -> bool:
    """Check if the given type field is subclass of ``Model``."""
    from ..abc import (
        Model,
    )

    return issubclass(type_field, Model)


def is_type_subclass(type_field: Any) -> bool:
    """Check if the given type field is subclass of ``type``."""
    return issubclass(type(type_field), type(type))


def is_aggregate_type(type_field: Any) -> bool:
    """Check if the given type is follows the ``Aggregate`` protocol."""
    return (
        is_type_subclass(type_field)
        and (is_model_subclass(type_field) or isinstance(type_field, ModelType))
        and {"uuid": UUID, "version": int}.items() <= type_field.type_hints.items()
    )


logger = logging.getLogger(__name__)

T = TypeVar("T")
K = TypeVar("K")


class TypeHintComparator(Generic[T, K]):
    """Type Hint Comparator class."""

    def __init__(self, first: Type[T], second: Type[K]):
        self._first = first
        self._second = second

    def match(self) -> bool:
        """Check if the two types match.

        :return: ``True`` if there is a match or ``False`` otherwise.
        """

        return self._compare(self._first, self._second)

    def _compare(self, first: Type[T], second: Type[K]) -> bool:
        if get_origin(first) is ModelRef:
            first = Union[(*get_args(first), UUID)]

        if get_origin(second) is ModelRef:
            second = Union[(*get_args(second), UUID)]

        if is_type_subclass(first) and is_model_subclass(first):
            first = first.model_type

        if is_type_subclass(second) and is_model_subclass(second):
            second = second.model_type

        if first == second:
            return True

        first_origin, second_origin = get_origin(first), get_origin(second)
        if first_origin is not None and self._compare(first_origin, second_origin):
            return self._compare_args(first, second)

        return False

    def _compare_args(self, first: Type[T], second: Type[K]) -> bool:
        first_args, second_args = get_args(first), get_args(second)
        if len(first_args) != len(second_args):
            return False
        return all(self._compare(fi, si) for fi, si in zip(first_args, second_args))



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
)

from .exceptions import (
    MinosException,
)
from .model import (
    AvroDataEncoder,
)


class Request(ABC):
    """Request interface."""

    @abstractmethod
    async def content(self, **kwargs) -> Any:
        """Get the request content.

        :param kwargs: Additional named arguments.
        :return: A list of instances.
        """
        raise NotImplementedError

    @abstractmethod
    def __eq__(self, other: Request) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError


class Response:
    """Response definition."""

    __slots__ = "_data"

    def __init__(self, data: Any):
        self._data = data

    # noinspection PyUnusedLocal
    async def content(self, **kwargs) -> Any:
        """Response content.

        :param kwargs: Additional named arguments.
        :return: A list of items.
        """
        return self._data

    # noinspection PyUnusedLocal
    async def raw_content(self, **kwargs) -> Any:
        """Raw response content.

        :param kwargs: Additional named arguments.
        :return: A list of raw items.
        """
        return AvroDataEncoder(self._data).build()

    def __eq__(self, other: Response) -> bool:
        return type(self) == type(other) and self._data == other._data

    def __repr__(self) -> str:
        return f"{type(self).__name__}({self._data!r})"


class ResponseException(MinosException):
    """Response Exception class."""



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .abc import (
    Model,
)
from .declarative import (
    Aggregate,
    AggregateDiff,
    AggregateRef,
    Command,
    CommandReply,
    CommandStatus,
    DeclarativeModel,
    Event,
    MinosModel,
    ValueObject,
)
from .dynamic import (
    BucketModel,
    DataTransferObject,
    DynamicModel,
    FieldsDiff,
)
from .fields import (
    Field,
    ModelField,
)
from .serializers import (
    AvroDataDecoder,
    AvroDataEncoder,
    AvroSchemaDecoder,
    AvroSchemaEncoder,
)
from .types import (
    Decimal,
    Enum,
    Fixed,
    MissingSentinel,
    ModelRef,
    ModelRefExtractor,
    ModelRefInjector,
    ModelType,
    NoneType,
    TypeHintBuilder,
    TypeHintComparator,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from dataclasses import (
    dataclass,
    field,
)
from typing import (
    Any,
    Generic,
    Optional,
    TypeVar,
)

T = TypeVar("T")
NoneType = type(None)


class MissingSentinel(Generic[T]):
    """Class to detect when a field is not initialized."""


@dataclass
class Fixed(Generic[T]):
    """
    Represents an Avro Fixed type
    size (int): Specifying the number of bytes per value
    """

    size: int
    default: Any = field(default=MissingSentinel)
    namespace: Optional[str] = None
    aliases: Optional[list[Any]] = None
    _dataclasses_custom_type: str = "Fixed"

    def __repr__(self) -> str:
        return f"Fixed(size={self.size})"


@dataclass
class Enum(Generic[T]):
    """
    Represents an Avro Enum type
    symbols (typing.List): Specifying the possible values for the enum
    """

    symbols: list[Any]
    default: Any = field(default=MissingSentinel)
    namespace: Optional[str] = None
    aliases: Optional[list[Any]] = None
    docs: Optional[str] = None
    _dataclasses_custom_type: str = "Enum"

    def __repr__(self) -> str:
        return f"Enum(symbols={self.symbols})"


@dataclass
class Decimal(Generic[T]):
    """
    Represents an Avro Decimal type
    precision (int): Specifying the number precision
    scale(int): Specifying the number scale. Default 0
    """

    precision: int
    scale: int = 0
    default: Any = field(default=MissingSentinel)
    _dataclasses_custom_type: str = "Decimal"

    # Decimal serializes to bytes, which doesn't support namespace
    aliases: Optional[list[Any]] = None

    def __repr__(self) -> str:
        return f"Decimal(precision={self.precision}, scale={self.scale})"



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from abc import (
    ABC,
    abstractmethod,
)
from typing import (
    Any,
    NoReturn,
)

from .setup import (
    MinosSetup,
)


class MinosBroker(ABC, MinosSetup):
    """Base Broker class."""

    @abstractmethod
    async def send(self, data: Any, **kwargs) -> NoReturn:
        """Send a list of ``Aggregate`` instances.

        :param data: The data to be send.
        :param kwargs: Additional named arguments.
        :return: This method does not return anything.
        """
        raise NotImplementedError



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from .abc import (
    DynamicModel,
)
from .bucket import (
    BucketModel,
)
from .diff import (
    FieldsDiff,
)
from .dto import (
    DataTransferObject,
)



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

from typing import (
    TYPE_CHECKING,
    NoReturn,
    Optional,
    Type,
)
from uuid import (
    UUID,
)

from dependency_injector.wiring import (
    Provide,
)

from ...configuration import (
    MinosConfig,
)
from ...exceptions import (
    MinosPreviousVersionSnapshotException,
    MinosRepositoryNotProvidedException,
)
from ...importlib import (
    import_module,
)
from ...repository import (
    MinosRepository,
    RepositoryAction,
    RepositoryEntry,
)
from ..entries import (
    SnapshotEntry,
)
from .abc import (
    PostgreSqlSnapshotSetup,
)

if TYPE_CHECKING:
    from ...model import (
        Aggregate,
        AggregateDiff,
    )


class PostgreSqlSnapshotBuilder(PostgreSqlSnapshotSetup):
    """Minos Snapshot Dispatcher class."""

    _repository: MinosRepository = Provide["repository"]

    def __init__(self, *args, repository: Optional[MinosRepository] = None, **kwargs):
        super().__init__(*args, **kwargs)
        if repository is not None:
            self._repository = repository

        if self._repository is None or isinstance(self._repository, Provide):
            raise MinosRepositoryNotProvidedException("A repository instance is required.")

    @classmethod
    def _from_config(cls, *args, config: MinosConfig, **kwargs) -> PostgreSqlSnapshotBuilder:
        return cls(*args, **config.snapshot._asdict(), **kwargs)

    async def are_synced(self, aggregate_name: str, aggregate_uuids: list[UUID]) -> bool:
        """Check if the snapshot has the latest version of a list of aggregates.

        :param aggregate_name: Class name of the ``Aggregate`` to be checked.
        :param aggregate_uuids: List of aggregate identifiers to be checked.

        :return: ``True`` if has the latest version for all the identifiers or ``False`` otherwise.
        """
        for aggregate_uuid in aggregate_uuids:
            if not await self.is_synced(aggregate_name, aggregate_uuid):
                return False
        return True

    async def is_synced(self, aggregate_name: str, aggregate_uuid: UUID) -> bool:
        """Check if the snapshot has the latest version of an ``Aggregate`` instance.

        :param aggregate_name: Class name of the ``Aggregate`` to be checked.
        :param aggregate_uuid: Identifier of the ``Aggregate`` instance to be checked.
        :return: ``True`` if has the latest version for the identifier or ``False`` otherwise.
        """
        query = self._repository.select(
            id_ge=await self._load_offset(), aggregate_name=aggregate_name, aggregate_uuid=aggregate_uuid
        )
        async for _ in query:
            return False
        return True

    async def dispatch(self) -> NoReturn:
        """Perform a dispatching step, based on the sequence of non already processed ``RepositoryEntry`` objects.

        :return: This method does not return anything.
        """
        offset = await self._load_offset()

        ids = set()

        def _update_offset(e: RepositoryEntry, o: int):
            ids.add(e.id)
            while o + 1 in ids:
                o += 1
                ids.remove(o)
            return o

        async for entry in self._repository.select(id_gt=offset):
            try:
                await self._dispatch_one(entry)
            except MinosPreviousVersionSnapshotException:
                pass
            offset = _update_offset(entry, offset)

        await self._store_offset(offset)

    async def _load_offset(self) -> int:
        # noinspection PyBroadException
        try:
            raw = await self.submit_query_and_fetchone(_SELECT_OFFSET_QUERY)
            return raw[0]
        except Exception:
            return 0

    async def _store_offset(self, offset: int) -> NoReturn:
        await self.submit_query(_INSERT_OFFSET_QUERY, {"value": offset})

    async def _dispatch_one(self, event_entry: RepositoryEntry) -> Optional[SnapshotEntry]:
        if event_entry.action is RepositoryAction.DELETE:
            return await self._submit_delete(event_entry)

        instance = await self._build_instance(event_entry)
        return await self._submit_instance(instance)

    async def _submit_delete(self, entry: RepositoryEntry) -> NoReturn:
        params = {
            "aggregate_uuid": entry.aggregate_uuid,
            "aggregate_name": entry.aggregate_name,
            "version": entry.version,
            "data": None,
        }
        await self.submit_query_and_fetchone(_INSERT_ONE_SNAPSHOT_ENTRY_QUERY, params)

    async def _build_instance(self, event_entry: RepositoryEntry) -> Aggregate:
        # noinspection PyTypeChecker
        diff = event_entry.aggregate_diff
        instance = await self._update_if_exists(diff)
        return instance

    async def _update_if_exists(self, diff: AggregateDiff) -> Aggregate:
        # noinspection PyBroadException
        try:
            # noinspection PyTypeChecker
            previous = await self._select_one_aggregate(diff.uuid, diff.name)
        except Exception:
            # noinspection PyTypeChecker
            aggregate_cls: Type[Aggregate] = import_module(diff.name)
            return aggregate_cls.from_diff(diff)

        if previous.version >= diff.version:
            raise MinosPreviousVersionSnapshotException(previous, diff)

        previous.apply_diff(diff)
        return previous

    async def _select_one_aggregate(self, aggregate_uuid: UUID, aggregate_name: str) -> Aggregate:
        snapshot_entry = await self._select_one(aggregate_uuid, aggregate_name)
        return snapshot_entry.aggregate

    async def _select_one(self, aggregate_uuid: UUID, aggregate_name: str) -> SnapshotEntry:
        raw = await self.submit_query_and_fetchone(_SELECT_ONE_SNAPSHOT_ENTRY_QUERY, (aggregate_uuid, aggregate_name))
        return SnapshotEntry(aggregate_uuid, aggregate_name, *raw)

    async def _submit_instance(self, aggregate: Aggregate) -> SnapshotEntry:
        snapshot_entry = SnapshotEntry.from_aggregate(aggregate)
        snapshot_entry = await self._submit_update_or_create(snapshot_entry)
        return snapshot_entry

    async def _submit_update_or_create(self, entry: SnapshotEntry) -> SnapshotEntry:
        params = {
            "aggregate_uuid": entry.aggregate_uuid,
            "aggregate_name": entry.aggregate_name,
            "version": entry.version,
            "data": entry.data,
        }
        response = await self.submit_query_and_fetchone(_INSERT_ONE_SNAPSHOT_ENTRY_QUERY, params)

        entry.created_at, entry.updated_at = response

        return entry


_SELECT_ONE_SNAPSHOT_ENTRY_QUERY = """
SELECT version, data, created_at, updated_at
FROM snapshot
WHERE aggregate_uuid = %s and aggregate_name = %s;
""".strip()

_INSERT_ONE_SNAPSHOT_ENTRY_QUERY = """
INSERT INTO snapshot (aggregate_uuid, aggregate_name, version, data, created_at, updated_at)
VALUES (
    %(aggregate_uuid)s,
    %(aggregate_name)s,
    %(version)s,
    %(data)s,
    default,
    default
)
ON CONFLICT (aggregate_uuid, aggregate_name)
DO
   UPDATE SET version = %(version)s, data = %(data)s, updated_at = NOW()
RETURNING created_at, updated_at;
""".strip()

_SELECT_OFFSET_QUERY = """
SELECT value
FROM snapshot_aux_offset
WHERE id = TRUE;
"""

_INSERT_OFFSET_QUERY = """
INSERT INTO snapshot_aux_offset (id, value)
VALUES (TRUE, %(value)s)
ON CONFLICT (id)
DO UPDATE SET value = %(value)s;
""".strip()



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from uuid import (
    UUID,
)

from ..abc import (
    DeclarativeModel,
)

logger = logging.getLogger(__name__)


class AggregateRef(DeclarativeModel):
    """Sub Aggregate class."""

    uuid: UUID
    version: int



"""
Copyright (C) 2021 Clariteia SL

This file is part of minos framework.

Minos framework can not be copied and/or distributed without the express permission of Clariteia SL.
"""
from __future__ import (
    annotations,
)

import logging
from typing import (
    AsyncIterator,
    Generic,
    NoReturn,
    Optional,
    TypeVar,
)
from uuid import (
    UUID,
)

from dependency_injector.wiring import (
    Provide,
)

from ....constants import (
    NULL_UUID,
)
from ....exceptions import (
    MinosBrokerNotProvidedException,
    MinosRepositoryManuallySetAggregateIdentifierException,
    MinosRepositoryManuallySetAggregateVersionException,
    MinosRepositoryNotProvidedException,
    MinosSnapshotNotProvidedException,
)
from ....networks import (
    MinosBroker,
)
from ....repository import (
    MinosRepository,
)
from ....snapshot import (
    MinosSnapshot,
)
from ..abc import (
    DeclarativeModel,
)
from .diff import (
    AggregateDiff,
)

T = TypeVar("T")
logger = logging.getLogger(__name__)


class Aggregate(DeclarativeModel, Generic[T]):
    """Base aggregate class."""

    uuid: UUID
    version: int

    _broker: MinosBroker = Provide["event_broker"]
    _repository: MinosRepository = Provide["repository"]
    _snapshot: MinosSnapshot = Provide["snapshot"]

    def __init__(
        self,
        *args,
        uuid: UUID = NULL_UUID,
        version: int = 0,
        _broker: Optional[MinosBroker] = None,
        _repository: Optional[MinosRepository] = None,
        _snapshot: Optional[MinosSnapshot] = None,
        **kwargs,
    ):

        super().__init__(uuid, version, *args, **kwargs)

        if _broker is not None:
            self._broker = _broker
        if _repository is not None:
            self._repository = _repository
        if _snapshot is not None:
            self._snapshot = _snapshot

        if self._broker is None or isinstance(self._broker, Provide):
            raise MinosBrokerNotProvidedException("A broker instance is required.")
        if self._repository is None or isinstance(self._repository, Provide):
            raise MinosRepositoryNotProvidedException("A repository instance is required.")
        if self._snapshot is None or isinstance(self._snapshot, Provide):
            raise MinosSnapshotNotProvidedException("A snapshot instance is required.")

    @classmethod
    async def get(
        cls,
        uuids: list[UUID],
        _broker: Optional[MinosBroker] = None,
        _repository: Optional[MinosRepository] = None,
        _snapshot: Optional[MinosSnapshot] = None,
    ) -> AsyncIterator[T]:
        """Get a sequence of aggregates based on a list of identifiers.

        :param uuids: list of identifiers.
        :param _broker: Broker to be set to the aggregates.
        :param _repository: Repository to be set to the aggregate.
        :param _snapshot: Snapshot to be set to the aggregate.
        :return: A list of aggregate instances.
        """

        if _broker is None:
            _broker = cls._broker
            if isinstance(_broker, Provide):
                raise MinosBrokerNotProvidedException("A broker instance is required.")

        if _repository is None:
            _repository = cls._repository
            if isinstance(_repository, Provide):
                raise MinosRepositoryNotProvidedException("A repository instance is required.")

        if _snapshot is None:
            _snapshot = cls._snapshot
            if isinstance(_snapshot, Provide):
                raise MinosSnapshotNotProvidedException("A snapshot instance is required.")

        # noinspection PyTypeChecker
        iterable = _snapshot.get(cls.classname, uuids, _broker=_broker, _repository=_repository, _snapshot=_snapshot)

        # noinspection PyTypeChecker
        async for aggregate in iterable:
            yield aggregate

    @classmethod
    async def get_one(
        cls,
        uuids: UUID,
        _broker: Optional[MinosBroker] = None,
        _repository: Optional[MinosRepository] = None,
        _snapshot: Optional[MinosSnapshot] = None,
    ) -> T:
        """Get one aggregate based on an identifier.

        :param uuids: Identifier of the aggregate.
        :param _broker: Broker to be set to the aggregates.
        :param _repository: Repository to be set to the aggregate.
        :param _snapshot: Snapshot to be set to the aggregate.
        :return: A list of aggregate instances.
        :return: An aggregate instance.
        """
        return await cls.get([uuids], _broker=_broker, _repository=_repository, _snapshot=_snapshot).__anext__()

    @classmethod
    async def create(
        cls, *args, _broker: Optional[MinosBroker] = None, _repository: Optional[MinosRepository] = None, **kwargs
    ) -> T:
        """Create a new ``Aggregate`` instance.

        :param args: Additional positional arguments.
        :param _broker: Broker to be set to the aggregates.
        :param _repository: Repository to be set to the aggregate.
        :param kwargs: Additional named arguments.
        :return: A new ``Aggregate`` instance.
        """
        if "uuid" in kwargs:
            raise MinosRepositoryManuallySetAggregateIdentifierException(
                f"The identifier must be computed internally on the repository. Obtained: {kwargs['uuid']}"
            )

        if "version" in kwargs:
            raise MinosRepositoryManuallySetAggregateVersionException(
                f"The version must be computed internally on the repository. Obtained: {kwargs['version']}"
            )

        instance = cls(*args, _broker=_broker, _repository=_repository, **kwargs)

        diff = AggregateDiff.from_aggregate(instance)
        entry = await instance._repository.create(diff)

        instance.uuid, instance.version = entry.aggregate_uuid, entry.version
        diff.uuid, diff.version = entry.aggregate_uuid, entry.version

        await instance._broker.send(diff, topic=f"{type(instance).__name__}Created")

        return instance

    # noinspection PyMethodParameters,PyShadowingBuiltins
    async def update(self, **kwargs) -> T:
        """Update an existing ``Aggregate`` instance.

        :param kwargs: Additional named arguments.
        :return: An updated ``Aggregate``  instance.
        """
        if "version" in kwargs:
            raise MinosRepositoryManuallySetAggregateVersionException(
                f"The version must be computed internally on the repository. Obtained: {kwargs['version']}"
            )

        # Update model...
        for key, value in kwargs.items():
            setattr(self, key, value)

        previous = await self.get_one(
            self.uuid, _broker=self._broker, _repository=self._repository, _snapshot=self._snapshot
        )
        diff = AggregateDiff.from_difference(self, previous)
        entry = await self._repository.update(diff)

        self.uuid, self.version = entry.aggregate_uuid, entry.version
        diff.uuid, diff.version = entry.aggregate_uuid, entry.version

        await self._broker.send(diff, topic=f"{type(self).__name__}Updated")

        return self

    async def save(self) -> NoReturn:
        """Store the current instance on the repository.

        If didn't exist previously creates a new one, otherwise updates the existing one.
        """
        is_creation = self.uuid == NULL_UUID
        if is_creation != (self.version == 0):
            if is_creation:
                raise MinosRepositoryManuallySetAggregateVersionException(
                    f"The version must be computed internally on the repository. Obtained: {self.version}"
                )
            else:
                raise MinosRepositoryManuallySetAggregateIdentifierException(
                    f"The uuid must be computed internally on the repository. Obtained: {self.uuid}"
                )

        if is_creation:
            new = await self.create(
                **{k: field.value for k, field in self.fields.items() if k not in ("uuid", "version")},
                _broker=self._broker,
                _repository=self._repository,
                _snapshot=self._snapshot,
            )
            self._fields |= new.fields
        else:
            await self.update(
                **{k: field.value for k, field in self.fields.items() if k not in ("uuid", "version")},
                _broker=self._broker,
                _repository=self._repository,
                _snapshot=self._snapshot,
            )

    async def refresh(self) -> NoReturn:
        """Refresh the state of the given instance.

        :return: This method does not return anything.
        """
        new = await type(self).get_one(
            self.uuid, _broker=self._broker, _repository=self._repository, _snapshot=self._snapshot
        )
        self._fields |= new.fields

    async def delete(self) -> NoReturn:
        """Delete the given aggregate instance.

        :return: This method does not return anything.
        """
        diff = AggregateDiff.from_deleted_aggregate(self)
        entry = await self._repository.delete(diff)

        self.uuid, self.version = entry.aggregate_uuid, entry.version
        diff.uuid, diff.version = entry.aggregate_uuid, entry.version

        await self._broker.send(diff, topic=f"{type(self).__name__}Deleted")

    def diff(self, another: Aggregate) -> AggregateDiff:
        """Compute the difference with another aggregate.

        Both ``Aggregate`` instances (``self`` and ``another``) must share the same ``uuid`` value.

        :param another: Another ``Aggregate`` instance.
        :return: An ``FieldsDiff`` instance.
        """
        return AggregateDiff.from_difference(self, another)

    def apply_diff(self, difference: AggregateDiff) -> NoReturn:
        """Apply the differences over the instance.

        :param difference: The ``FieldsDiff`` containing the values to be set.
        :return: This method does not return anything.
        """
        if self.uuid != difference.uuid:
            raise ValueError(
                f"To apply the difference, it must have same uuid. "
                f"Expected: {self.uuid!r} Obtained: {difference.uuid!r}"
            )
        logger.debug(f"Applying {difference!r} to {self!r}...")
        for field in difference.fields_diff:
            setattr(self, field.name, field.value)
        self.version = difference.version

    @classmethod
    def from_diff(cls, difference: AggregateDiff, *args, **kwargs) -> T:
        """Build a new instance from an ``AggregateDiff``.

        :param difference: The difference that contains the data.
        :param args: Additional positional arguments.
        :param kwargs: Additional named arguments.
        :return: A new ``Aggregate`` instance.
        """
        return cls(*args, uuid=difference.uuid, version=difference.version, **difference.fields_diff, **kwargs)
